#
#    IniFile.pm - perl module to mess with .ini files (regutils package)
#    Copyright (C) 1998 Memorial University of Newfoundland
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#    
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#    

#
# This file contains the following:
#    package IniFile;
#    package IniFile::Field;
#    package IniFile::Key;
#    package IniFile::Filter;
#

#
# Creates the following data structure:
#	{
#	    'nextOrderId' => some-number,
#	    'keys' => {
#		    {
#			'name' => "...",
#			'orderId' => some number,
#			'fields' => [
#				{
#				    'name' => "..."
#				    'value' => "...",
#				    'deleted' => 0 or 1,
#				    'annotation' => "...",
#				},
#				...
#			    ],
#			'deleted' => 0 or 1,
#			'annotation' => "...",
#		    },
#		    ...
#		},
#	}
#
# 'orderId' is used to preserve the order in which keys appear in a file.
# 'annotation' is used to decorate inidiff output.
# 'deleted' is used when generating, writing, reading, and applying patch files.
#
#

package IniFile;

use vars qw($VERSION);
$VERSION = '0.8';

use strict;
use Carp;

use vars qw(&new &write &findKey &keys &removekey &addKey);
use vars qw($errorString);

{
    package IniFile::Field;

    use strict;
    use Carp;

    use vars qw(&new &name &canonName &value &deleted &annotation
		&setFrom &write &canonicalize );


    sub new
    {
	my $proto = shift;
	confess "new IniFile::Field takes 4 arguments" if @_ != 4;
	my($name, $value, $del, $annotation) = @_;

	my $field = {
		    'name' => $name,
		    'canonName' => canonicalize($name),
		    'value' => $value,
		    'deleted' => $del,
		    'annotation' => $annotation,
		};

	bless $field, (ref($proto) || $proto);
	return $field;
    }

    sub name {
	my $field = shift;
	if (@_) { $field->{'name'} = shift }
	return $field->{'name'};
    }
    sub canonName {
	my $field = shift;
	if (@_) { $field->{'canonName'} = shift }
	return $field->{'canonName'};
    }
    sub value {
	my $field = shift;
	if (@_) { $field->{'value'} = shift }
	return $field->{'value'};
    }
    sub deleted {
	my $field = shift;
	if (@_) { $field->{'deleted'} = shift }
	return $field->{'deleted'};
    }
    sub annotation {
	my $field = shift;
	if (@_) { $field->{'annotation'} = shift }
	return $field->{'annotation'};
    }

    sub setFrom
    {
	my $field = shift;
	my $from = shift;
	$field->name($from->name);
	$field->value($from->value);
	$field->deleted($from->deleted);
	$field->annotation($from->annotation);
	return $field;
    }

    sub canonicalize
    {
	# Called as object method
	if (@_ > 0 && ref $_[0]) {
	    return ${$_[0]}->canonName if (@_ == 1);
	}
	# Called as object or class method with argument
	if (@_ == 2) {
	    shift;
	} elsif (@_ != 1) {
	    confess "wrong number of args" if (@_ != 0);
	}
	my $name = $_[0];
	$name =~ tr/A-Z/a-z/;
	return $name;
    }

    sub write
    {
	my($field, $fileHandle) = @_;

	print $fileHandle "; ", $field->annotation, "\r\n"
	    if defined $field->annotation;
	print $fileHandle $field->name;
	if ($field->deleted) {
	    print $fileHandle "-";
	} elsif (defined $field->value) {
	    print $fileHandle "=", $field->value;
	}
	print $fileHandle "\r\n";
    }
}

{
    package IniFile::Key;

    use strict;
    use Carp;
    use IO::File;

    use vars qw(&new &name &canonName &orderId &deleted &annotation &fields
		&canonicalize &findField &addField &appendField &removeField
		&write
		);

    sub new
    {
	my $proto = shift;
	confess "new IniFile::Key takes 3 arguments" if @_ != 3;
	my($name, $del, $annotation) = @_;

	my $key = { 'name' => $name,
		    'canonName' => canonicalize($name),
		    'orderId' => undef,
		    'deleted' => $del,
		    'annotation' => $annotation,
		    'fields' => [],
		  };

	bless $key, (ref($proto) || $proto);
	return $key;
    }

    sub name {
	my $key = shift;
	# Do not change canonName (see kludge in IniFile::new regarding patches)
	if (@_) { $key->{'name'} = shift }
	return $key->{'name'};
    }
    sub canonName {
	my $key = shift;
	if (@_) { $key->{'canonName'} = shift }
	return $key->{'canonName'};
    }
    sub orderId {
	my $key = shift;
	if (@_) { $key->{'orderId'} = shift }
	return $key->{'orderId'};
    }
    sub deleted {
	my $key = shift;
	if (@_) { $key->{'deleted'} = shift }
	return $key->{'deleted'};
    }
    sub annotation {
	my $key = shift;
	if (@_) { $key->{'annotation'} = shift }
	return $key->{'annotation'};
    }
    sub fields {
	my $key = shift;
	confess "too many args" if @_ > 0;
	return $key->{'fields'};
    }

    sub canonicalize
    {
	# Called as object method
	if (@_ > 0 && ref $_[0]) {
	    return ${$_[0]}->canonName if (@_ == 1);
	}
	# Called as object or class method with argument
	if (@_ == 2) {
	    shift;
	} elsif (@_ != 1) {
	    confess "wrong number of args" if (@_ != 0);
	}
	my $name = $_[0];
	$name =~ tr/A-Z/a-z/;
	return $name;
    }

    # Note: only finds first field...
    sub findField
    {
	my($key, $fieldName) = @_;
	$fieldName = IniFile::Field->canonicalize($fieldName);

	my($field);
	foreach $field (@{$key->fields}) {
	    return $field if ($field->canonName eq $fieldName);
	}
	return undef;
    }

    sub addField
    {
	my $key = shift;
	my $field = shift;

	my $xfield = $key->findField($field->name);
	if (defined $xfield) {
	    $xfield->setFrom($field);
	    return $xfield;
	}
	return $key->appendField($field);
    }

    sub appendField
    {
	my $key = shift;
	my $field = shift;

	push(@{$key->fields}, $field);
	return $field;
    }

    #
    # Remove a field, either by name or by reference
    #
    sub removeField
    {
	my $found = 0;
	my($key, $arg) = @_;

	if (ref $arg) {
	    my $fieldToast = $arg;
	    for (my $i = 0; $i < @{$key->fields}; $i++) {
		my $field = ${$key->fields}[$i];
		if ($field eq $fieldToast) {
		    splice(@{$key->fields}, $i, 1);
		    $i--;
		    $found++;
		}
	    }
	} else {
	    my $fieldName = $arg;
	    for (my $i = 0; $i < @{$key->fields}; $i++) {
		my $field = ${$key->fields}[$i];
		if ($field->canonName eq $fieldName) {
		    splice(@{$key->fields}, $i, 1);
		    $i--;
		    $found++;
		}
	    }
	}
	return $found;
    }


    sub write
    {
	my($key, $fileHandle) = @_;
	my($del) = $key->deleted ? '-' : '';

	print $fileHandle "; ", $key->annotation, "\r\n"
	    if defined $key->annotation;
	print $fileHandle "[", $key->name, "]", $del, "\r\n";
	if (!$key->deleted) {
	    my $field;
	    foreach $field (@{$key->fields}) {
		$field->write($fileHandle);
	    }
	}
	print $fileHandle "\r\n";
    }
}


$errorString = undef;


sub new
{
    my $proto = shift;

    my $ini = {
	    # Used to generate monotonically increasing key ids - used
	    # to reserve order of ini file.
	    'nextOrderId' => 0,
	    'keys' => {},
	};

    bless $ini, (ref($proto) || $proto);

    return $ini if (@_ == 0);
    confess "new called with too many arguments" if (@_ > 2);

    my($fileHandle, $isPatch) = @_;	# Patches: allow duplicate key names
    my($key);
    my $ok = 1;

    $isPatch = 0 if !defined $isPatch;

    while (<$fileHandle>) {
	chomp;
	#
	# Strip comments - not in key names ([...]) and not in strings..
	#
	if (/^\s*(\[[^]]+])(.*)$/) {
	    my($key, $rest) = ($1, $2);
	    $rest =~ s/;.*//;
	    $_ = $key . $rest;
	} elsif (!/^\s*;/ && /"/) {
	    # Slow, but perl won't go exponential...
	    my $line = '';
	    while (/^([^";]*"[^"]*")(.*)/) {
		$line .= $1;
		$_ = $2;
	    }
	    # Some (burnt) ini files have unmatched quotes... rather
	    # that toast these, we assume they have no comments.
	    if (!/"/) {
		s/;.*//;
	    }
	    $_ = $line . $_;
	} else {
	    s/;.*//;
	}
	s/\s*$//;		# includes \r...
	next if /^$/;		# skip blank lines
	if (/^\s*\[([^]]+)](-?)$/) {
	    my($name, $del) = ($1, $2);
	    if ($isPatch) {
		# Patches are a bit strange as there can be duplicate
		# key names - to deal with this, the canonName (hash index)
		# is a generated (unique) thing and we fix up the real name
		# after the key is created.
		$key = $ini->addKey(new IniFile::Key("[$ini->{'nextOrderId'}]",
						$del eq '-', undef));
		$key->name($name);
	    } else {
		if ($del eq '-') {
		    $errorString = "$.: non-patch file has deleted key";
		    $ok = 0;
		    last;
		}
		if ($ini->findKey($name)) {
		    $errorString = "$.: duplicate key: $name";
		    $ok = 0;
		    last;
		}
		$key = $ini->addKey(new IniFile::Key($name, $del eq '-', undef));
	    }
	    next;
	}

	my($name, $value);
	if (/"/) {
	    # Slow, but perl won't go exponential...
	    my $line = '';
	    while (/^([^"=]*"[^"]*")(.*)$/) {
		$line .= $1;
		$_ = $2;
	    }
	    if (/^([^"]*)=(.*)$/) {
		$name = $line . $1;
		$value = $2;
	    } elsif (/"/) {
		# Wonder if this will be a problem...
		$errorString = "$.: unmatched quote (no preceeding =)";
		$ok = 0;
		last;
	    } else {
		$name = $line . $_;
		$value = undef;
	    }
	} elsif (/^([^=]+)=(.*)$/) {
	    $name = $1;
	    $value = $2;
	} else {
	    $name = $_;
	    $value = undef;
	}
	my $del = 0;
	if (!defined $value && $name =~ /-$/) {
	    if (!$isPatch) {
		$errorString = "$.: non-patch file has deleted field";
		$ok = 0;
		last;
	    }
	    $del = 1;
	    chop $name;
	}
	if (!defined $key) {
	    $errorString = "$.: field outside of key\n";
	    $ok = 0;
	    last;
	}
	if ($key->deleted) {
	    $errorString = "$.: deleted key has field\n";
	    $ok = 0;
	    last;
	}
	$key->appendField(new IniFile::Field($name, $value, $del, undef));
    }
    return undef if !$ok;
    return $ini;
}

sub write
{
    my($ini, $fileHandle) = @_;
    my($key);

    foreach $key (@{$ini->keys}) {
	$key->write($fileHandle);
    }
}

sub findKey
{
    my($ini, $keyName) = @_;
    return $ini->{'keys'}->{IniFile::Key->canonicalize($keyName)};
}

sub keys
{
    my($ini) = @_;

    return [ sort { $a->orderId <=> $b->orderId }
		  values(%{$ini->{'keys'}}) ];
}

#
# Remove a key, either by name or by reference
#
sub removeKey
{
    my($ini, $arg) = @_;

    if (ref $arg) {
	my $keyToast = $arg;
	my ($name, $key);
	while (($name, $key) = each %{$ini->{'keys'}}) {
	    if ($key eq $keyToast) {
		delete $ini->{'keys'}->{$name};
	    }
	}
    } else {
	my $keyName = $arg;
	return delete $ini->{'keys'}->{IniFile::Key->canonicalize($keyName)};
    }
}

sub addKey
{
    my($ini, $key) = @_;

    $ini->{'keys'}->{$key->canonName} = $key;
    $key->orderId($ini->{'nextOrderId'}++);
    return $key;
}

{
    package IniFile::Filter;

    use strict;
    use Carp;

    use vars qw(&new &readConf &filter);
    use vars qw($errorString);

    sub new
    {
	my $proto = shift;
	confess "new IniFile::Filter takes no arguments" if @_ != 0;

	my $field = {
		    'keyFilters' => []
		};

	bless $field, (ref($proto) || $proto);
	return $field;
    }

    # This a RegFile::Filter are very (very) similar - keep them that way
    sub readConf
    {
	my($filter, $file) = @_;

	my($keyActions) = undef;
	my($entryActions) = undef;

	my $in = new IO::File "$file", "r";
	if (!defined $in) {
	    $errorString = "can't open $file - $!";
	    return 0;
	}

	while (<$in>) {
	    next if (/^\s*(#|$)/);
	    # Trim whitespace
	    s/^\s+//; s/\s+$//;

	    # include another filter file?
	    if (/^\s*include\s+"[^"]*"\s*$/) {
		my($ifile) = ($1);

		# End the previous key (with error checking)
		if (defined $keyActions && !@$keyActions) {
		    $errorString = "$file:$.: previous key has no actions";
		    $in->close;
		    return 0;
		} elsif (defined $entryActions && !@$entryActions) {
		    $errorString = "$file:$.: previous name/value pattern has no actions";
		    $in->close;
		    return 0;
		}
		$keyActions = $entryActions = undef;

		# If not an absolute path, try relative to this file first.
		my($mypath) = $file;
		$mypath =~ s:/+[^/]*$::;
		if ($ifile !~ /^\// && $mypath ne '' && $mypath ne $file
		    && -e "$mypath/$ifile")
		{
		    return 0 if (!$filter->readConf("$mypath/$ifile"));
		} else {
		    return 0 if (!$filter->readConf($ifile));
		}
		next;
	    }

	    # A new key?
	    if (/^\[(.+)](|\s*-)$/) {
		my($keyPat, $isDel) = ($1, $2 eq '' ? 0 : 1);

		if (defined $keyActions && !@$keyActions) {
		    $errorString = "$file:$.: previous key has no actions";
		    $in->close;
		    return 0;
		} elsif (defined $entryActions && !@$entryActions) {
		    $errorString = "$file:$.: previous name/value pattern has no actions";
		    $in->close;
		    return 0;
		}

		push(@{$filter->{'keyFilters'}}, {
			'keyPat' => $1,
			'deleteAll' => $isDel,
			'keyActions' => [ ]
		    });
		$keyActions = $isDel ? undef : ${$filter}[$#$filter]->{'keyActions'};
		next;
	    }
	    # A new entry match operator?
	    if (/^\s*(name|value)\s+(\S.*)$/) {
		my($matchOn, $matchPat) = ($1, $2);
		if (!defined $keyActions) {
		    $errorString = "$file:$.: name/value pattern found outside key";
		    $in->close;
		    return 0;
		}
		if (defined $entryActions && !@$entryActions) {
		    $errorString = "$file:$.: previous name/value pattern has no actions";
		    $in->close;
		    return 0;
		}
		$entryActions = [];
		push(@$keyActions, { 'matchOn' => $matchOn,
				    'matchPat' => $matchPat,
				    'entryActions' => $entryActions,
				    });
		next;
	    }
	    # A entry substitution
	    if (/^\s*subst\s+(name|value)\s+(\S.*)$/) {
		my($action, $subst) = ($1, $2);
		if (!defined $keyActions) {
		    $errorString = "$file:$.: substitution found outside key";
		    $in->close;
		    return 0;
		}
		if (!defined $entryActions) {
		    $errorString = "$file:$.: substitution found outside entry/key pattern";
		    $in->close;
		    return 0;
		}
		push(@$entryActions, { 'action' => "subst\u$action",
				       'subst' => $subst,
				       });
		next;
	    }
	    # A entry deletion
	    if (/^\s*delete$/) {
		if (!defined $keyActions) {
		    $errorString = "$file:$.: delete entry found outside key";
		    $in->close;
		    return 0;
		}
		if (!defined $entryActions) {
		    $errorString = "$file:$.: delete entry found outside entry/key pattern";
		    $in->close;
		    return 0;
		}
		push(@$entryActions, { 'action' => 'delete',
				       'subst' => undef,
				       });
		next;
	    }
	    $errorString = "$file:$.: unexpected line";
	    $in->close;
	    return 0;
	}
	$in->close;
	return 1;
    }

    #
    # Given a ini object, modify it by applying the filtering commands
    # (deletions and substitutions) contained in the filter object.
    # Returns true iff there are no problems.
    #
    sub filter
    {
	my($filter, $ini) = @_;

	my $key;
	foreach $key ($ini->keys) {
	    my $keyFilt;
	    foreach $keyFilt (@{$filter->{'keyFilters'}}) {
		next if ($key->name !~ /^$keyFilt->{'keyPat'}$/i);
		if ($keyFilt->{'deleteAll'}) {
		    $ini->removeKey($key);
		    next;	# no point in going on...
		}
	    }
	    # Must be entry substituions/deletions
	    my $keyAction;
	    foreach $keyAction (@{$keyFilt->{'keyActions'}}) {
		my $field;
		foreach $field ($key->fields) {
		    my $target = $keyAction->{'matchOn'} eq 'name'
				    ? $field->name : $field->value;
		    next if ($target !~ /^$keyAction->{'matchPat'}$/i);
		    #
		    # Have a match - carry out entry actions
		    #
		    my $entryAction;
		    foreach $entryAction (@{$keyAction->{'entryActions'}}) {
			if ($entryAction->{'action'} eq 'substName') {
			    my $name = $field->name;
			    eval "\$name =~ $keyAction->{'subst'}";
			    if ($@ ne '') {
				$errorString = "error substituting $target using $keyAction->{'subst'}";
				return undef;
			    }
			    $field->name($name);
			} elsif ($entryAction->{'action'} eq 'substValue') {
			    my $value = $field->value;
			    eval "\$value =~ $keyAction->{'subst'}";
			    if ($@ ne '') {
				$errorString = "error substituting $target using $keyAction->{'subst'}";
				return undef;
			    }
			    $field->value($value);
			} elsif ($entryAction->{'action'} eq 'delete') {
			    $key->removeField($field);
			    last;
			} else {
			    $errorString = "regfilter::filter: internal error - unknown entry action: $entryAction->{'action'}";
			    return undef;
			}
		    }
		}
	    }
	}

	return 1;
    }
}

1;
